@@ -1,137 +1,141 @@
-#############################################################################
 # pristine-lfs
 #
 # store pristine tarballs in Git LFS
 #
-# Copyright (C) 2019 Collabora Ltd
+# Copyright (C) 2019—2021 Collabora Ltd
 # Andrej Shadura <andrew.shadura@collabora.co.uk>
+#
+# SPDX-License-Identifier: GPL-2.0-or-later
 
-# This program is free software; you can redistribute it
-# and/or modify it under the terms of the GNU General Public
-# License as published by the Free Software Foundation; either
-# version 2 of the License, or (at your option) any later
-# version.
-
-# This program is distributed in the hope that it will be
-# useful, but WITHOUT ANY WARRANTY; without even the implied
-# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
-# PURPOSE.  See the GNU General Public License version 2
-# text for more details.
-
-# You should have received a copy of the GNU General Public
-# License along with this package; if not, write to the Free
-# Software Foundation, Inc., 51 Franklin St, Fifth Floor,
-# Boston, MA  02110-1301 USA
-#############################################################################
-
-from gettext import gettext as _
+import argparse
+import logging
 import os
 import sys
-from . import util
-from .util import *
-import logging
-import argparse
-import sh
 from fnmatch import fnmatch
+from gettext import gettext as _
 from pathlib import Path
+from typing import IO, Optional, Union
 
+import sh
 from debian import deb822
-from debian.changelog import Version, Changelog
+from debian.changelog import Changelog, Version
 
-from typing import Optional, List
+from .util import (
+    Abort,
+    check_branch,
+    checkout_lfs_file,
+    checkout_package,
+    commit_lfs_file,
+    commit_lfs_files,
+    find_branch,
+    find_remote_branches,
+    list_git_files,
+    track_remote_branch,
+)
 
-def find_branch(branch: str) -> str:
+
+def do_commit(tarball: IO[bytes], branch: str, message: Optional[str] = None, force_overwrite: bool = False, **kwargs):
+    """
+    Commit the open file to a branch using Git LFS.
+    Set force_overwrite to overwrite an existing file with the same name and a different checksum.
+    Message may contain "%s" which gets replaced with a comma-separate list of the file committed.
+    """
     if check_branch(branch) is None:
-        remote_branches = find_remote_branches(branch)
-        if remote_branches:
-            commit, branch = remote_branches[0]
-        else:
-            raise util.Abort(_('No branch {branch} found, not even among remote branches').format(branch=branch))
-    return branch
+        if find_remote_branches(branch):
+            track_remote_branch(branch)
+    commit_lfs_file(tarball, branch, message, overwrite=force_overwrite)
 
-def do_commit(args: argparse.Namespace):
-    if check_branch(args.branch) is None:
-        if find_remote_branches(args.branch):
-            track_remote_branch(args.branch)
-    commit_lfs_file(args.tarball, args.branch, args.message, overwrite=args.force_overwrite)
 
-def do_checkout(args: argparse.Namespace):
-    branch = find_branch(args.branch)
-    os.makedirs(args.outdir, exist_ok=True)
-    if args.auto:
+def do_checkout(branch: str, tarball: Optional[str] = None, outdir: Union[str, Path] = '.', full: bool = False, **kwargs):
+    """
+    Check out one or multiple files.
+    If tarball is non-None:
+        * file name only: tarball to check out to outdir.
+        * path with to file: the location where to check out to
+    If tarball is None:
+        * a tarball corresponding to the latest entry in
+          debian/changelog is found and checked out
+
+    When full is set, tarball is a name of a .dsc file;
+    all corresponding tarball *and* the .dsc file are
+    checked out. If None, debian/changelog is used
+    to find the set of files to check out.
+    """
+    branch = find_branch(branch)
+    if tarball:
+        path, slash, tarball = tarball.rpartition('/')
+        if path:
+            outdir = path
+    else:
         changelog = Path("debian/changelog")
-        if not changelog.is_file():
-            raise util.Abort(_('file %s not found ') % changelog)
         with changelog.open() as f:
             ch = Changelog(f, max_blocks=1)
         package, version = ch.package, ch.version
-    if args.full:
-        if not args.auto:
-            dsc_file = args.tarball
+
+    outdir = Path(outdir)
+    outdir.mkdir(parents=True, exist_ok=True)
+
+    if full:
+        if tarball:
+            dsc_file = tarball
         else:
             fver = Version(ch.version)
             fver.epoch = None
             dsc_file = f'{package}_{fver}.dsc'
-        logging.info(_("Checking out file {} in {}").format(dsc_file, args.outdir))
-        checkout_lfs_file(branch, dsc_file, args.outdir)
+        logging.info(_("Checking out file {} in {}").format(dsc_file, outdir))
+        checkout_lfs_file(branch, dsc_file, outdir)
         if dsc_file.endswith('.dsc'):
-            with (Path(args.outdir) / dsc_file).open('r') as dsc:
+            with (outdir / dsc_file).open('r') as dsc:
                 d = deb822.Dsc(dsc)
             package = d['Source']
             version = Version(d['Version'])
             files = [f["name"] for f in d["Files"]]
-            checkout_package(package, version, branch, args.outdir, files)
+            checkout_package(package, version, branch, outdir, files)
     else:
-        if not args.auto:
-            logging.info(_("Checking out file {} in {}").format(args.tarball, args.outdir))
-            checkout_lfs_file(branch, args.tarball, args.outdir)
+        if tarball:
+            logging.info(_("Checking out file {} in {}").format(tarball, outdir))
+            checkout_lfs_file(branch, tarball, outdir)
         else:
-            checkout_package(package, version, branch, args.outdir)
+            checkout_package(package, version, branch, outdir)
 
-def checkout_package(package: str, version: Version, branch: str, outdir: str, requested: Optional[Sequence[str]] = None):
-        logging.info(_("Checking out files for {package} version {version} to {outdir}:").format(package=package, version=version, outdir=outdir))
-        tarball_glob = f'{package}_{version.upstream_version}.orig.tar.*'
-        component_tarball_glob = f'{package}_{version.upstream_version}.orig-*.tar.*'
 
-        files = list_git_files(branch)
-        if requested:
-            # TODO: handle missing files
-            tarballs = [f for f in files if f in requested]
-        else:
-            tarballs = [f for f in files if fnmatch(f, tarball_glob) or fnmatch(f, component_tarball_glob)]
-
-        os.makedirs(outdir, exist_ok=True)
-        for f in tarballs:
-            logging.info("         ... {}".format(f))
-            checkout_lfs_file(branch, f, outdir)
-        logging.info(_("Done."))
-
-def do_list(args: argparse.Namespace):
-    branch = find_branch(args.branch)
+def do_list(branch: str, **kwargs):
+    """
+    List all files on the specified branch except .gitattributes.
+    """
+    branch = find_branch(branch)
     for f in list_git_files(branch):
         if f != '.gitattributes':
             print(f)
 
-def do_import(args: argparse.Namespace):
-    d = deb822.Dsc(args.dsc)
+
+def do_import(dsc: IO[str], branch: str, message: Optional[str] = None, force_overwrite: bool = False, full: bool = False, **kwargs):
+    """
+    Import all tarballs and detached signatures related to an open .dsc file.
+    Set force_overwrite to overwrite an existing file with the same name and a different checksum.
+    Message may contain "%s" which gets replaced with a comma-separate list of the file committed.
+    When full is set, the .dsc file and the Debian packaging tarball are also imported.
+    """
+    d = deb822.Dsc(dsc)
     package = d['Source']
     version = Version(d['Version'])
 
     tarball_glob = f'{package}_{version.upstream_version}.orig.tar.*'
     component_tarball_glob = f'{package}_{version.upstream_version}.orig-*.tar.*'
-    dsc_dir = os.path.dirname(args.dsc.name)
+    dsc_dir = os.path.dirname(dsc.name)
 
-    if check_branch(args.branch) is None:
-        if find_remote_branches(args.branch):
-            track_remote_branch(args.branch)
+    if check_branch(branch) is None:
+        if find_remote_branches(branch):
+            track_remote_branch(branch)
 
-    tarballs = [os.path.join(dsc_dir, f['name']) for f in d['Files'] if args.full or fnmatch(f['name'], tarball_glob) or fnmatch(f['name'], component_tarball_glob)]
-    if args.full:
-        tarballs += [args.dsc.name]
+    tarballs = [os.path.join(dsc_dir, f['name']) for f in d['Files'] if full or fnmatch(f['name'], tarball_glob) or fnmatch(f['name'], component_tarball_glob)]
+    if full:
+        tarballs += [dsc.name]
 
     if tarballs:
         logging.info("Importing: %s" % " ".join(tarballs))
-        commit_lfs_files([open(tarball, 'rb') for tarball in tarballs], args.branch, args.message, overwrite=args.force_overwrite)
+        commit_lfs_files([open(tarball, 'rb') for tarball in tarballs], branch, message, overwrite=force_overwrite)
+
 
 def main():
     prog = os.path.basename(sys.argv[0])
@@ -139,7 +143,7 @@ def main():
     parser = argparse.ArgumentParser(description=_('store pristine tarballs in Git LFS'), prog=prog)
     parser.add_argument('-v', '--verbose', action='count', help=_('be more verbose'))
     parser.add_argument('--debug', action='store_const', const=2, dest='verbose', help=_('be debuggingly verbose'))
-    parser.set_defaults(verbose=0, func=lambda x: parser.print_usage(file=sys.stderr))
+    parser.set_defaults(verbose=0, func=lambda *x, **kw: parser.print_usage(file=sys.stderr))
     subparsers = parser.add_subparsers(required=False)
 
     parser_commit = subparsers.add_parser('commit', help=_('commit a tarball'))
@@ -180,15 +184,18 @@ def main():
     logging.getLogger(sh.__name__).setLevel(logging.WARNING - 10 * (args.verbose - 1))
 
     try:
-        args.func(args)
+        args.func(**vars(args))
     except sh.ErrorReturnCode as e:
         print(_('Failed to run %s:') % e.full_cmd, file=sys.stderr)
         print(e.stderr.decode(sh.DEFAULT_ENCODING, "replace"), file=sys.stderr)
         sys.exit(e.exit_code)
-    except KeyboardInterrupt as e:
+    except FileNotFoundError as e:
+        print(_('abort: file %s not found ') % e.filename, file=sys.stderr)
+        sys.exit(1)
+    except KeyboardInterrupt:
         print(file=sys.stderr)
         print(_('about: Interrupted by user'), file=sys.stderr)
         sys.exit(1)
-    except util.Abort as e:
+    except Abort as e:
         print(_("abort: %s\n") % e, file=sys.stderr)
         sys.exit(1)
