@@ -1,40 +1,28 @@
-#############################################################################
 # pristine-lfs
 #
 # Git and Git LFS routines
 # This requires Git and git-lfs to be installed.
 #
-# Copyright (C) 2019 Collabora Ltd
+# Copyright (C) 2019—2021 Collabora Ltd
 # Andrej Shadura <andrew.shadura@collabora.co.uk>
+#
+# SPDX-License-Identifier: GPL-2.0-or-later
 
-# This program is free software; you can redistribute it
-# and/or modify it under the terms of the GNU General Public
-# License as published by the Free Software Foundation; either
-# version 2 of the License, or (at your option) any later
-# version.
+from __future__ import annotations
 
-# This program is distributed in the hope that it will be
-# useful, but WITHOUT ANY WARRANTY; without even the implied
-# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
-# PURPOSE.  See the GNU General Public License version 2
-# text for more details.
-
-# You should have received a copy of the GNU General Public
-# License along with this package; if not, write to the Free
-# Software Foundation, Inc., 51 Franklin St, Fifth Floor,
-# Boston, MA  02110-1301 USA
-#############################################################################
-
-from gettext import gettext as _
 import logging
-import sh
-import sh.contrib
 import os
-from pathlib import Path
-from fnmatch import fnmatchcase
 from contextlib import contextmanager
+from fnmatch import fnmatch, fnmatchcase
+from gettext import gettext as _
+from pathlib import Path
+from typing import IO, Any, Generator, Iterable, Mapping, Optional, Sequence, Union, cast
+
+import sh
+from debian.changelog import Version
+
+from .gitwrap import git
 
-from typing import Optional, Tuple, List, Sequence, IO, Any, Union, Mapping
 
 gitattributes = """*.tar.* filter=lfs diff=lfs merge=lfs -text
 *.tar.*.asc -filter -diff merge=binary -text
@@ -53,16 +41,9 @@ case "$GIT_LFS_SKIP_PUSH" in
 esac
 """
 
-def git(*args, index: Optional[Path] = None, **kwargs):
-    e = os.environ
-    if index:
-        e["GIT_INDEX_FILE"] = str(index)
-    return sh.contrib.git(*args, **kwargs, _env=e)
-
-git.lfs = sh.contrib.git.lfs
 
 @contextmanager
-def open_index(name: str) -> Path:
+def open_index(name: str) -> Generator[Path, None, None]:
     index = Path(git_dir()) / f"index-{name}"
     if index.exists():
         index.unlink()
@@ -72,11 +53,14 @@ def open_index(name: str) -> Path:
         if index.exists():
             index.unlink()
 
-AttributeValue = Union[str, bool, None]
 
-def parse_attr(attr: str) -> Tuple[str, AttributeValue]:
+AttributeValue = Union[str, bool, None]
+Attribute = tuple[str, AttributeValue]
+
+
+def parse_attr(attr: str) -> Attribute:
     """
-    Parse a git attribute into its value:
+    Parse a Git attribute into its value:
 
     >>> parse_attr('attr')
     ('attr', True)
@@ -93,18 +77,63 @@ def parse_attr(attr: str) -> Tuple[str, AttributeValue]:
         return attr[1:], False
     if '=' not in attr:
         return attr, True
-    return tuple(attr.split('='))
+    return cast(Attribute, tuple(attr.split('=', maxsplit=1)))
 
-def parse_git_attributes(s: str) -> List[Tuple[str, Mapping[str, AttributeValue]]]:
-   lines = [l.strip() for l in s.splitlines()]
-   lines = [l.split() for l in lines if len(l) and not l.startswith('#')]
-   return [(l[0], {k: v for k, v in [parse_attr(a) for a in l[1:]]}) for l in lines if len(l)]
 
-default_gitattributes = parse_git_attributes(gitattributes)
+def parse_git_attributes(s: str) -> Iterable[tuple[str, Mapping[str, AttributeValue]]]:
+    """
+    Parse Git attributes from a string:
+
+    >>> list(parse_git_attributes('''
+    ... ab*     merge=filfre
+    ... abc     -foo -bar
+    ... *.c     frotz
+    ... *
+    ... '''))
+    [('ab*', {'merge': 'filfre'}),
+     ('abc', {'foo': False, 'bar': False}),
+     ('*.c', {'frotz': True}),
+     ('*', {})]
+    >>> list(parse_git_attributes('[attr]binary -diff -merge -text'))
+    [('[attr]binary', {'diff': False, 'merge': False, 'text': False})]
+    >>> list(parse_git_attributes('''
+    ... # this is a comment
+    ... a*      foo !bar -baz
+    ... abc     foo bar baz
+    ... *.jpg -text -diff
+    ... *               text=auto
+    ... *.txt           text
+    ... *.vcproj        text eol=crlf
+    ... *.sh            text eol=lf
+    ... *.jpg           -text
+    ... '''))
+    [('a*', {'foo': True, 'bar': None, 'baz': False}),
+     ('abc', {'foo': True, 'bar': True, 'baz': True}),
+     ('*.jpg', {'text': False, 'diff': False}),
+     ('*', {'text': 'auto'}),
+     ('*.txt', {'text': True}),
+     ('*.vcproj', {'text': True, 'eol': 'crlf'}),
+     ('*.sh', {'text': True, 'eol': 'lf'}),
+     ('*.jpg', {'text': False})]
+    """
+    lines = [line.strip() for line in s.splitlines()]
+    # strip comments, split each line by consecutive spaces
+    lines = [line.split() for line in lines if len(line) and not line.startswith('#')]
+    for line in lines:
+        if len(line):
+            glob, *attrs = line
+            yield glob, {k: v for k, v in [
+                parse_attr(a) for a in attrs
+            ]}
+
+
+default_gitattributes = list(parse_git_attributes(gitattributes))
+
 
 class Abort(Exception):
     pass
 
+
 def check_branch(name: str) -> Optional[str]:
     """
     Check a branch exists, return the hash it points at, if it does.
@@ -116,38 +145,76 @@ def check_branch(name: str) -> Optional[str]:
     except sh.ErrorReturnCode:
         return None
 
+
 def git_dir() -> str:
     return git('rev-parse', '--git-dir').strip('\n')
 
+
 def git_head() -> str:
     return git('rev-parse', '-q', '--verify', '--symbolic-full-name', 'HEAD', _ok_code=[0, 1]).strip('\n')
 
-def find_remote_branches(name: str) -> List[Tuple[str, str]]:
+
+def branch_remote(name: str) -> Optional[str]:
+    '''
+    Find an upstream remote for a tracking branch
+    '''
+    if not name.startswith('refs/heads/') and not name.startswith('refs/'):
+        name = 'refs/heads/' + name
+    return git('for-each-ref', '--format=%(upstream:remotename)', '--', name, _ok_code=[0, 1]).strip('\n') or None
+
+
+def find_remote_branches(name: str) -> list[tuple[str, str]]:
     try:
-        branches = [l.split(' ') for l in git('show-ref', '--', name).splitlines()]
+        branches = [line.split(' ') for line in git('show-ref', '--', name).splitlines()]
         return [(b[0], b[1]) for b in branches if b[1].startswith('refs/remotes/')]
     except sh.ErrorReturnCode:
         return []
 
+
+def preferred_remote_branch(remote_branches: list[tuple[str, str]]) -> tuple[str, str]:
+    logging.debug("Remote branches: %r", remote_branches)
+    current_remote = branch_remote(git_head())
+    logging.debug("Current remote: %r", current_remote)
+    remote_branches = [
+        (commit, ref) for (commit, ref) in remote_branches
+        if current_remote and ref.startswith('refs/remotes/' + current_remote)
+    ] + remote_branches
+    return remote_branches[0]
+
+
 def track_remote_branch(name: str):
     remote_branches = find_remote_branches(name)
     if len(remote_branches) == 0:
         raise RuntimeError('remote branch expected but not found')
-    commit, branch = remote_branches[0]
+    commit, branch = preferred_remote_branch(remote_branches)
     git('branch', '--track', name, branch)
 
+
+def find_branch(branch: str) -> str:
+    if check_branch(branch) is None:
+        remote_branches = find_remote_branches(branch)
+        if remote_branches:
+            commit, branch = preferred_remote_branch(remote_branches)
+        else:
+            raise Abort(_('No branch {branch} found, not even among remote branches').format(branch=branch))
+    return branch
+
+
 def store_lfs_object(io: Any) -> str:
     return str(git.lfs.clean(io.name, _in=io))
 
+
 def store_git_object(io: Any) -> str:
     return git('hash-object', '-w', '--stdin', _in=io).strip('\n')
 
+
 def stage_file(filename: Union[str, bytes], io: Any, index: Path = None):
     blob = store_git_object(io)
     if isinstance(filename, bytes):
         filename = filename.decode()
     git('update-index', '--add', '--replace', '--cacheinfo', "100644,%s,%s" % (blob, filename), index=index)
 
+
 def create_commit(branch: str, message: str, index: Path = None) -> str:
     tree = git('write-tree', index=index).strip('\n')
     if not len(tree):
@@ -164,7 +231,34 @@ def create_commit(branch: str, message: str, index: Path = None) -> str:
 
     return commit
 
+
+def refresh_main_index():
+    git('update-index', '--ignore-submodules', '-q', '--ignore-missing', '--unmerged', '--refresh')
+
+
 def parse_diff_entry(entry: str) -> Mapping[str, str]:
+    r"""
+    >>> parse_diff_entry(
+    ... ":100644 100644 777f41fb18215a23a779a831b38d24ff6171775a c5ac094096d35c687c0cc5054b5a1433c293ebd5 M\tpristine_lfs/main.py"
+    ... )
+    {'srcmode': '100644',
+     'dstmode': '100644',
+     'srchash': '777f41fb18215a23a779a831b38d24ff6171775a',
+     'dsthash': 'c5ac094096d35c687c0cc5054b5a1433c293ebd5',
+     'srcname': 'pristine_lfs/main.py',
+     'dstname': '',
+     'status': 'M'}
+    >>> parse_diff_entry(
+    ... ":100644 100644 f0adb461cd94419245cf0b16a0edf1e2daf6688b f0adb461cd94419245cf0b16a0edf1e2daf6688b R100\told\tnew"
+    ... )
+    {'srcmode': '100644',
+     'dstmode': '100644',
+     'srchash': 'f0adb461cd94419245cf0b16a0edf1e2daf6688b',
+     'dsthash': 'f0adb461cd94419245cf0b16a0edf1e2daf6688b',
+     'srcname': 'old',
+     'dstname': 'new',
+     'status': 'R100'}
+    """
     treediff, names = entry.split('\t', maxsplit=1)
     srcmode, dstmode, srchash, dsthash, status = treediff.lstrip(':').split(' ')
     srcname, _, dstname = names.partition('\t')
@@ -178,12 +272,14 @@ def parse_diff_entry(entry: str) -> Mapping[str, str]:
         'status': status,
     }
 
+
 def commit_lfs_file(io: IO[bytes], branch: str, template: str = None, overwrite: bool = False):
     """
     Store the file in the LFS storage and commit it to a branch.
     """
     commit_lfs_files([io], branch)
 
+
 def commit_lfs_files(ios: Sequence[IO[bytes]], branch: str, template: str = None, overwrite: bool = False):
     """
     Store the files in the LFS storage and commit them to a branch.
@@ -215,7 +311,7 @@ def commit_lfs_files(ios: Sequence[IO[bytes]], branch: str, template: str = None
                 stage_file(filename, metadata, index=index)
 
         if check_branch(branch) is not None:
-            diff = git('diff-index', '--cached', branch).strip().splitlines()
+            diff = git('diff-index', '--cached', branch, index=index).strip().splitlines()
             if not diff:
                 logging.info(_("Nothing to commit"))
                 return
@@ -227,27 +323,29 @@ def commit_lfs_files(ios: Sequence[IO[bytes]], branch: str, template: str = None
         if not template:
             template = "pristine-lfs data for %s"
 
-        message = template % ', '.join([os.path.basename(io.name) for io in ios])
+        message = template.replace('%s', ', '.join([os.path.basename(io.name) for io in ios]))
 
-        commit = create_commit(branch, message, index=index)
+        create_commit(branch, message, index=index)
 
-        # if the branch is currently checked out, reset it
-        if git_head() == f'refs/heads/{branch}':
-            git('reset', '--hard')
+        refresh_main_index()
 
-def list_lfs_files(branch: str) -> List[str]:
+
+def list_lfs_files(branch: str) -> list[str]:
     return git.lfs('ls-files', '--name-only', branch).splitlines()
 
-def parse_entry(entry: str) -> Tuple[str, ...]:
+
+def parse_entry(entry: str) -> tuple[str, ...]:
     info, name = entry.split('\t')
     mode, type, hash = info.split(' ')
     return mode, type, hash, name
 
+
 def list_git_files(branch: str) -> Mapping[str, str]:
-    entries = [parse_entry(l) for l in git('ls-tree', '-r', '--full-name', branch).splitlines()]
+    entries = [parse_entry(line) for line in git('ls-tree', '-r', '--full-name', branch).splitlines()]
     return {e[3]: e[2] for e in entries if e[1] == 'blob'}
 
-def is_lfs_managed(filename: str, attributes: List[Tuple[str, Mapping[str, AttributeValue]]]):
+
+def is_lfs_managed(filename: str, attributes: Iterable[tuple[str, Mapping[str, AttributeValue]]]):
     lfs_managed = False
     for pattern, attrs in attributes:
         if fnmatchcase(filename, pattern):
@@ -255,7 +353,8 @@ def is_lfs_managed(filename: str, attributes: List[Tuple[str, Mapping[str, Attri
                 lfs_managed = attrs['filter'] == 'lfs'
     return lfs_managed
 
-def checkout_lfs_file(branch: str, filename: str, outdir: str = '.'):
+
+def checkout_lfs_file(branch: str, filename: str, outdir: Union[str, Path] = '.'):
     files = list_git_files(branch)
     if '.gitattributes' in files:
         attributes = parse_git_attributes(git('cat-file', 'blob', files['.gitattributes']))
@@ -271,3 +370,21 @@ def checkout_lfs_file(branch: str, filename: str, outdir: str = '.'):
             git.lfs.smudge(filename, _out=tarball, _in=metadata)
         else:
             git('cat-file', 'blob', files[filename], _out=tarball)
+
+
+def checkout_package(package: str, version: Version, branch: str, outdir: Union[str, Path], requested: Optional[Sequence[str]] = None):
+    logging.info(_("Checking out files for {package} version {version} to {outdir}:").format(package=package, version=version, outdir=outdir))
+    tarball_glob = f'{package}_{version.upstream_version}.orig.tar.*'
+    component_tarball_glob = f'{package}_{version.upstream_version}.orig-*.tar.*'
+
+    files = list_git_files(branch)
+    if requested:
+        # TODO: handle missing files
+        tarballs = [f for f in files if f in requested]
+    else:
+        tarballs = [f for f in files if fnmatch(f, tarball_glob) or fnmatch(f, component_tarball_glob)]
+
+    for f in tarballs:
+        logging.info("         ... {}".format(f))
+        checkout_lfs_file(branch, f, outdir)
+    logging.info(_("Done."))
